#!/usr/bin/env python3
"""
Generates python client from Swagger specification.
"""

# pylint: disable=missing-docstring,too-many-instance-attributes,too-many-locals,too-many-ancestors,too-many-branches
# pylint: disable=too-many-statements,too-many-lines

from typing import MutableMapping, Union, Set, List, TextIO, Optional  # pylint: disable=unused-import

import collections

import swagger_to
import swagger_to.intermediate
import swagger_to.swagger


class Typedef:
    """
    Represents a python type.
    """

    def __init__(self) -> None:
        self.description = ''
        self.identifier = ''

    def __str__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, self.identifier)


class Booldef(Typedef):
    pass


class Intdef(Typedef):
    pass


class Floatdef(Typedef):
    pass


class Strdef(Typedef):
    pass


class Bytesdef(Typedef):
    pass


class Filedef(Typedef):
    """
    Represents a file type in form-data parameters.
    """
    pass


class Attribute:
    def __init__(self) -> None:
        self.typedef = None  # type: Optional[Typedef]
        self.description = ''
        self.name = ''
        self.required = False

        self.classdef = None  # type: Optional[Classdef]


class Classdef(Typedef):
    def __init__(self) -> None:
        super().__init__()
        self.attributes = collections.OrderedDict()  # type: MutableMapping[str, Attribute]


class Listdef(Typedef):
    def __init__(self) -> None:
        super().__init__()

        self.items = None  # type: Union[None, Typedef]


class Dictdef(Typedef):
    def __init__(self) -> None:
        super().__init__()

        self.values = None  # type: Union[None, Typedef]


class Parameter:
    """
    Represents a parameter of a request.
    """

    def __init__(self) -> None:
        self.name = ''
        self.typedef = None  # type: Optional[Typedef]
        self.required = False
        self.description = None  # type: Optional[str]


class Response:
    def __init__(self) -> None:
        self.code = ''
        self.typedef = None  # type: Optional[Typedef]
        self.description = ''


class Request:
    """
    Represents a request function of the client.
    """

    def __init__(self) -> None:
        self.operation_id = ''
        self.path = ''
        self.method = ''
        self.description = ''

        # the union of all parameters
        self.parameters = []  # type: List[Parameter]

        # parameters split into categories
        self.body_parameter = None  # type: Optional[Parameter]
        self.query_parameters = []  # type: List[Parameter]
        self.path_parameters = []  # type: List[Parameter]
        self.formdata_parameters = []  # type: List[Parameter]
        self.file_parameters = []  # type: List[Parameter]

        self.responses = collections.OrderedDict()  # type: MutableMapping[str, Response]
        self.produces = []  # type: List[str]


def anonymous_or_get_typedef(intermediate_typedef: swagger_to.intermediate.Typedef,
                             typedefs: MutableMapping[str, Typedef]) -> Typedef:
    """
    If the type has an identifier, it is retrieved from the translated typedefs.

    Otherwise, it is translated on the spot to the corresponding typescript type.

    :param intermediate_typedef: to be translated
    :param typedefs: translated type definitions
    :return: translated type
    """
    if intermediate_typedef.identifier:
        if intermediate_typedef.identifier not in typedefs:
            raise KeyError("Intermediate typedef not found in the translated typedefs: {!r}".format(
                intermediate_typedef.identifier))

        return typedefs[intermediate_typedef.identifier]

    return to_typedef(intermediate_typedef=intermediate_typedef)


def to_typedef(intermediate_typedef: swagger_to.intermediate.Typedef) -> Typedef:
    typedef = None  # type: Optional[Typedef]

    if isinstance(intermediate_typedef, swagger_to.intermediate.Primitivedef):
        if intermediate_typedef.type == 'boolean':
            typedef = Booldef()
        elif intermediate_typedef.type == 'integer':
            typedef = Intdef()
        elif intermediate_typedef.type == 'number':
            typedef = Floatdef()
        elif intermediate_typedef.type == 'string':
            typedef = Strdef()
        elif intermediate_typedef.type == 'file':
            typedef = Filedef()
        else:
            raise NotImplementedError("Converting intermediate type to Python is not supported: {}".format(
                intermediate_typedef.type))

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Arraydef):
        typedef = Listdef()
        typedef.items = to_typedef(intermediate_typedef=intermediate_typedef.items)

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Mapdef):
        typedef = Dictdef()
        typedef.values = to_typedef(intermediate_typedef=intermediate_typedef.values)

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Objectdef):
        typedef = Classdef()

        for intermediate_prop in intermediate_typedef.properties.values():
            attr = Attribute()
            attr.description = intermediate_prop.description
            attr.name = intermediate_prop.name
            attr.typedef = to_typedef(intermediate_typedef=intermediate_prop.typedef)
            attr.required = intermediate_prop.required
            attr.classdef = typedef

            typedef.attributes[attr.name] = attr

        typedef.attributes = collections.OrderedDict(
            sorted(list(typedef.attributes.items()), key=lambda kv: not kv[1].required))
    else:
        raise NotImplementedError("Converting intermediate typedef to Python is not supported: {!r}".format(
            type(intermediate_typedef)))

    typedef.description = intermediate_typedef.description
    typedef.identifier = intermediate_typedef.identifier

    assert typedef is not None

    return typedef


def to_typedefs(
        intermediate_typedefs: MutableMapping[str, swagger_to.intermediate.Typedef]) -> MutableMapping[str, Typedef]:
    typedefs = collections.OrderedDict()  # type: MutableMapping[str, Typedef]

    for intermediate_typedef in intermediate_typedefs.values():
        assert intermediate_typedef is not None

        typedef = to_typedef(intermediate_typedef=intermediate_typedef)
        typedefs[typedef.identifier] = typedef

    return typedefs


def to_parameter(intermediate_parameter: swagger_to.intermediate.Parameter,
                 typedefs: MutableMapping[str, Typedef]) -> Parameter:
    """
    Translates an intermediate parameter to a typescript parameter.

    :param intermediate_parameter: to be translated
    :param typedefs: translated type definitions
    :return: translated parameter
    """
    param = Parameter()
    param.name = intermediate_parameter.name
    param.typedef = anonymous_or_get_typedef(intermediate_typedef=intermediate_parameter.typedef, typedefs=typedefs)
    param.required = intermediate_parameter.required
    param.description = intermediate_parameter.description
    return param


def to_response(intermediate_response: swagger_to.intermediate.Response,
                typedefs: MutableMapping[str, Typedef]) -> Response:
    """
    Translates an intermediate response to a python response.

    :param intermediate_response: to be translated
    :param typedefs: translated type definitions
    :return: translated response
    """
    resp = Response()
    resp.code = intermediate_response.code
    resp.typedef = None if intermediate_response.typedef is None else \
        anonymous_or_get_typedef(intermediate_typedef=intermediate_response.typedef, typedefs=typedefs)
    resp.description = intermediate_response.description
    return resp


def to_request(endpoint: swagger_to.intermediate.Endpoint, typedefs: MutableMapping[str, Typedef]) -> Request:
    """
    Translates an endpoint to a client request function.

    :param endpoint: to be translated
    :param typedefs: translated type definitions
    :return: request function
    """

    req = Request()
    req.description = endpoint.description
    req.method = endpoint.method
    req.operation_id = endpoint.operation_id
    req.path = endpoint.path

    for intermediate_param in endpoint.parameters:
        param = to_parameter(intermediate_parameter=intermediate_param, typedefs=typedefs)

        if isinstance(param.typedef, Filedef):
            req.file_parameters.append(param)
        elif intermediate_param.in_what == 'body':
            if req.body_parameter is not None:
                raise KeyError("Duplicate body parameters in an endpoint: {} {}".format(
                    req.body_parameter.name, intermediate_param.name))

            req.body_parameter = param
        elif intermediate_param.in_what == 'query':
            req.query_parameters.append(param)
        elif intermediate_param.in_what == 'path':
            req.path_parameters.append(param)
        elif intermediate_param.in_what == 'formData':
            req.formdata_parameters.append(param)
        else:
            raise NotImplementedError("Unsupported parameter 'in' to Python translation: {}".format(
                intermediate_param.in_what))

        req.parameters.append(param)

    # parameters are sorted so that first come the required ones; python requires the optional parameters to come
    # at the end.
    req.parameters.sort(key=lambda param: not param.required)

    for code, intermediate_resp in endpoint.responses.items():
        req.responses[code] = to_response(intermediate_response=intermediate_resp, typedefs=typedefs)

    req.produces = endpoint.produces[:]

    return req


def to_requests(endpoints: List[swagger_to.intermediate.Endpoint],
                typedefs: MutableMapping[str, Typedef]) -> List[Request]:
    """
    Translates the endpoints to typescript request functions.

    Endpoints which do not produce strictly 'application/json' are ignored.

    :param endpoints: to be translated
    :param typedefs: translated type definitions
    :return: translated request functions
    """
    requests = []  # type: List[Request]
    for endpoint in endpoints:
        requests.append(to_request(endpoint=endpoint, typedefs=typedefs))

    return requests


INDENT = ' ' * 4


def write_imports(import_set: Set[str], fid: TextIO) -> None:
    import_lst = sorted(list(import_set))
    for import_stmt in sorted(import_lst):
        fid.write('import ' + '{}\n'.format(import_stmt))
    if len(import_lst) > 0:
        fid.write("\n\n")


def type_expression(typedef: Typedef, path: Optional[str] = None) -> str:
    """
    Translates the typedef to a type expression.

    :param typedef: to be translated
    :param path: path in the intermediate representation
    :return: type expression
    """
    # pylint: disable=too-many-return-statements
    if isinstance(typedef, Booldef):
        return 'bool'
    elif isinstance(typedef, Intdef):
        return 'int'
    elif isinstance(typedef, Floatdef):
        return 'float'
    elif isinstance(typedef, Strdef):
        return 'str'
    elif isinstance(typedef, Bytesdef):
        return 'bytes'
    elif isinstance(typedef, Filedef):
        return 'BinaryIO'
    elif isinstance(typedef, Listdef):
        return 'List[' + type_expression(typedef=typedef.items, path=str(path) + '.items') + "]"
    elif isinstance(typedef, Dictdef):
        return 'Dict[str, ' + type_expression(typedef=typedef.values, path=str(path) + '.values') + "]"
    elif isinstance(typedef, Classdef):
        if typedef.identifier == '':
            raise NotImplementedError(
                "Translating an anonymous class to a type expression is not supported: {}".format(path))

        return typedef.identifier
    else:
        raise NotImplementedError("Translating the typedef to a type expression is not supported: {!r}: {}".format(
            type(typedef), path))


def attribute_as_argument(attribute: Attribute) -> str:
    argtype = type_expression(typedef=attribute.typedef, path=attribute.classdef.identifier + "." + attribute.name)

    if not attribute.required:
        return '{}: Optional[{}] = None'.format(attribute.name, argtype)

    return '{}: {}'.format(attribute.name, argtype)


def write_header(service_name: str, fid: TextIO) -> None:
    fid.write('#!bin/bash/python3\n')
    fid.write('# Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!\n')
    fid.write('"""Implements the client for {}."""\n\n'.format(service_name))
    fid.write("# pylint: skip-file\n")
    fid.write("# pydocstyle: add-ignore=D105,D107,D401\n\n")

    fid.write('import contextlib\n')
    fid.write("from typing import Any, BinaryIO, Dict, List, Optional\n\n")
    fid.write("import requests\n")
    fid.write("import requests.auth\n")


def write_footer(fid: TextIO) -> None:
    fid.write('# Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!')


def write_comment(comment: str, indent: str, fid: TextIO) -> None:
    lines = comment.strip().splitlines()
    for i, line in enumerate(lines):
        rstripped = line.rstrip()
        if len(rstripped) > 0:
            fid.write(indent + '# {}'.format(rstripped))
        else:
            fid.write(indent + '#')

        if i < len(lines) - 1:
            fid.write('\n')


def write_docstring(docstring: str, indent: str, fid: TextIO) -> None:
    if not docstring:
        raise ValueError("Unexpected empty docstring")

    docstring = docstring[0].upper() + docstring[1:]
    docstring_lines = docstring.splitlines()

    if len(docstring_lines) == 1:
        fid.write(''.join([indent, '"""', docstring, '"""']))
        return

    fid.write(indent + '"""\n')
    for line in docstring_lines:
        if line.strip():
            fid.write(indent)
            fid.write(line)
        fid.write('\n')

    fid.write('\n')
    fid.write(indent + '"""')


def write_class(classdef: Classdef, fid: TextIO) -> None:
    if classdef.identifier == '':
        raise ValueError("Expected a classdef with an identifier, but got a classdef with an empty identifier.")

    fid.write("class {}:\n".format(classdef.identifier))
    if classdef.description:
        write_docstring(docstring=classdef.description, indent=INDENT, fid=fid)
        fid.write('\n\n')

    if not classdef.attributes:
        fid.write(INDENT + 'pass')
        return

    prefix = INDENT + 'def __init__(self'
    suffix = ') -> None:'

    args = []  # type: List[str]
    for attr in classdef.attributes.values():
        args.append(attribute_as_argument(attribute=attr))

    line = prefix + ', ' + ', '.join(args) + suffix
    if len(line) <= 80:
        fid.write(line)
    else:
        fid.write(prefix)
        for arg in args:
            fid.write(',\n')

            fid.write(' ' * (len(prefix) - len('self')) + arg)

        fid.write(suffix)

    fid.write('\n')

    for i, attr in enumerate(classdef.attributes.values()):
        if i > 0:
            fid.write('\n\n')

        if attr.description:
            write_comment(comment=attr.description, indent=INDENT * 2, fid=fid)
            fid.write('\n')
        fid.write(INDENT * 2 + 'self.{0} = {0}'.format(attr.name))

    fid.write('\n\n')
    fid.write(INDENT + 'def to_jsonable(self) -> Dict[str, Any]:\n')
    fid.write(INDENT * 2 + '"""\n')
    fid.write(INDENT * 2 + 'Dispatches the conversion to {}_to_jsonable.\n\n'.format(
        swagger_to.snake_case(identifier=classdef.identifier)))

    fid.write(INDENT * 2 + ':return: JSON-able representation\n\n')
    fid.write(INDENT * 2 + '"""\n')

    fid.write(INDENT * 2 + 'return {}_to_jsonable(self)'.format(swagger_to.snake_case(identifier=classdef.identifier)))


def default_attribute_value(typedef: Typedef) -> str:
    # pylint: disable=too-many-return-statements
    if isinstance(typedef, Booldef):
        return 'False'
    elif isinstance(typedef, Intdef):
        return '0'
    elif isinstance(typedef, Floatdef):
        return '0.0'
    elif isinstance(typedef, Strdef):
        return "''"
    elif isinstance(typedef, Bytesdef):
        return "b''"
    elif isinstance(typedef, Listdef):
        return "[]"
    elif isinstance(typedef, Dictdef):
        return "dict()"
    elif isinstance(typedef, Classdef):
        return "new_{}()".format(swagger_to.snake_case(identifier=typedef.identifier))
    else:
        raise NotImplementedError("Translating the typedef to a default value is not supported: {}".format(typedef))


def write_class_factory_method(classdef: Classdef, fid: TextIO) -> None:
    fid.write('def new_{}() -> {}:\n'.format(
        swagger_to.snake_case(identifier=classdef.identifier), classdef.identifier))

    fid.write(INDENT + '"""Generates a default instance of {}."""\n'.format(classdef.identifier))

    if not classdef.attributes:
        fid.write(INDENT + "return {}()".format(classdef.identifier))
        return

    prefix = INDENT + 'return {}('.format(classdef.identifier)
    suffix = ')'
    args = []  # type: List[str]
    for attr in classdef.attributes.values():
        if attr.required:
            args.append('{}={}'.format(attr.name, default_attribute_value(typedef=attr.typedef)))

    line = prefix + ', '.join(args) + suffix
    if len(line) <= 80:
        fid.write(line)
    else:
        fid.write(prefix)
        fid.write(args[0])

        for arg in args[1:]:
            fid.write(',\n')
            fid.write(' ' * len(prefix) + arg)

        fid.write(suffix)


def write_from_obj(classdefs: List[Classdef], fid: TextIO):
    # yapf: disable
    fid.write('''def from_obj(obj: Any, expected: List[type], path: str = '') -> Any:
    """
    Checks and converts the given obj along the expected types.

    :param obj: to be converted
    :param expected: list of types representing the (nested) structure
    :param path: to the object from the root object
    :return: the converted object

    """
    if not expected:
        raise ValueError("`expected` is empty, but at least one type needs to be specified.")

    exp = expected[0]

    if exp == float:
        if isinstance(obj, int):
            return float(obj)

        if isinstance(obj, float):
            return obj

        raise ValueError("Expected object of type int or float at {!r}, but got {}.".format(path, type(obj)))

    if exp in [bool, int, str, list, dict]:
        if not isinstance(obj, exp):
            raise ValueError("Expected object of type {} at {!r}, but got {}.".format(exp, path, type(obj)))

    if exp in [bool, int, float, str]:
        return obj

    if exp == list:
        lst = []  # type: List[Any]
        for i, value in enumerate(obj):
            lst.append(from_obj(value, expected=expected[1:], path=path + '[{}]'.format(i)))

        return lst

    if exp == dict:
        adict = dict()  # type: Dict[str, Any]
        for key, value in obj.items():
            if not isinstance(key, str):
                raise ValueError("Expected a key of type str at path {!r}, got: {}".format(path, type(key)))

            adict[key] = from_obj(value, expected=expected[1:], path=path + '["{}"]'.format(key))

        return adict'''.replace(' ' * 4, INDENT))

    for classdef in classdefs:
        fid.write('\n\n')
        fid.write(INDENT + 'if exp == {}:\n'.format(classdef.identifier))
        fid.write(INDENT * 2 + 'return {}_from_obj(obj, path=path)'.format(
            swagger_to.snake_case(classdef.identifier)))

    fid.write('\n\n')

    fid.write(INDENT + 'raise ValueError("Unexpected `expected` type: {}".format(exp))')


def expected_type_expression(typedef: Typedef) -> str:
    # pylint: disable=too-many-return-statements
    if isinstance(typedef, Booldef):
        return 'bool'
    elif isinstance(typedef, Intdef):
        return 'int'
    elif isinstance(typedef, Floatdef):
        return 'float'
    elif isinstance(typedef, Strdef):
        return "str"
    elif isinstance(typedef, Bytesdef):
        return "bytes"
    elif isinstance(typedef, Listdef):
        return "list, {}".format(expected_type_expression(typedef=typedef.items))
    elif isinstance(typedef, Dictdef):
        return "dict, {}".format(expected_type_expression(typedef=typedef.values))
    elif isinstance(typedef, Classdef):
        return typedef.identifier
    else:
        raise NotImplementedError(
            "Translating the typedef to an expected type is not supported: {}".format(typedef))


def write_class_from_obj(classdef: Classdef, fid: TextIO) -> None:
    fid.write('def {}_from_obj(obj: Any, path: str = "") -> {}:\n'.format(
        swagger_to.snake_case(identifier=classdef.identifier), classdef.identifier))

    fid.write(INDENT + '"""Generates an instance of {} from a dictionary object."""\n'.format(
        classdef.identifier))

    # yapf: disable
    fid.write(INDENT + 'if not isinstance(obj, dict):\n' +
              INDENT * 2 +
              'raise ValueError("Expected a dict at path {}, but got: {}".format(path, type(obj)))\n\n')
    fid.write(INDENT + 'for key in obj:\n' +
              INDENT * 2 + 'if not isinstance(key, str):\n' +
              INDENT * 3 +
              'raise ValueError("Expected a key of type str at path {}, but got: {}".format(path, type(key)))\n\n')
    # yapf: enable

    if not classdef.attributes:
        fid.write(INDENT + "return {}()".format(classdef.identifier))
        return

    def write_set_attr_stmt(indention: str, set_attr_stmt_parts: List[str]) -> None:
        prefix, value, expected, path, type_suffix = set_attr_stmt_parts

        line = indention + prefix + ', '.join([value, expected, path]) + type_suffix
        if len(line) <= 80:
            fid.write(line)
            return

        fid.write(indention + prefix + value + ',\n')
        fid.write(' ' * len(indention + prefix) + expected + ',\n')
        fid.write(' ' * len(indention + prefix) + path + type_suffix)

    for i, attr in enumerate(classdef.attributes.values()):
        if i > 0:
            fid.write('\n\n')

        attr_type_expr = type_expression(typedef=attr.typedef, path=attr.classdef.identifier + "." + attr.name)

        # yapf: disable
        set_attr_stmt_parts = [
            '{}_from_obj = from_obj('.format(attr.name),
            'obj["{0}"]'.format(attr.name),
            'expected=[{}]'.format(
                expected_type_expression(typedef=attr.typedef)),
            'path=path + ".{}")'.format(attr.name),
            '  # type: {}'.format(attr_type_expr)
        ]
        # yapf: enable

        if attr.required:
            write_set_attr_stmt(indention=INDENT, set_attr_stmt_parts=set_attr_stmt_parts)
        else:
            fid.write(INDENT + '''{}_from_obj = None  # type: Optional[{}]\n'''.format(attr.name, attr_type_expr))

            fid.write(INDENT + 'if "{}" in obj:\n'.format(attr.name))
            write_set_attr_stmt(indention=INDENT * 2, set_attr_stmt_parts=set_attr_stmt_parts)

    fid.write('\n\n')

    prefix = INDENT + 'return {}('.format(classdef.identifier)
    suffix = ')'
    args = []  # type: List[str]
    for attr in classdef.attributes.values():
        args.append('{0}={0}_from_obj'.format(attr.name))

    line = prefix + ', '.join(args) + suffix
    if len(line) <= 80:
        fid.write(line)
    else:
        fid.write(prefix)
        fid.write(args[0])

        for arg in args[1:]:
            fid.write(',\n')
            fid.write(' ' * len(prefix) + arg)

        fid.write(suffix)


def write_to_jsonable(classdefs: List[Classdef], fid: TextIO):
    # yapf: disable
    fid.write('''def to_jsonable(obj: Any, expected: List[type], path: str = "") -> Any:
    """
    Checks and converts the given object along the expected types to a JSON-able representation.

    :param obj: to be converted
    :param expected: list of types representing the (nested) structure
    :return: JSON-able representation of the object

    """
    if not expected:
        raise ValueError("`expected` is empty, but at least one type needs to be specified.")

    exp = expected[0]
    if not isinstance(obj, exp):
        raise ValueError("Expected object of type {} at path {!r}, but got {}.".format(exp, path, type(obj)))

    # Assert on primitive types to help type-hinting.
    if exp == bool:
        assert isinstance(obj, bool)
        return obj

    if exp == int:
        assert isinstance(obj, int)
        return obj

    if exp == float:
        assert isinstance(obj, float)
        return obj

    if exp == str:
        assert isinstance(obj, str)
        return obj

    if exp == list:
        assert isinstance(obj, list)

        lst = []  # type: List[Any]
        for i, value in enumerate(obj):
            lst.append(to_jsonable(value, expected=expected[1:], path=''.join([path, '[', str(i), ']'])))

        return lst

    if exp == dict:
        assert isinstance(obj, dict)

        adict = dict()  # type: Dict[str, Any]
        for key, value in obj.items():
            if not isinstance(key, str):
                raise ValueError("Expected a key of type str at path {!r}, got: {}".format(path, type(key)))

            adict[key] = to_jsonable(value, expected=expected[1:], path=''.join([path, '[', key, ']']))

        return adict'''.replace(' ' * 4, INDENT))

    for classdef in classdefs:
        fid.write('\n\n')
        fid.write(INDENT + 'if exp == {}:\n'.format(classdef.identifier))
        fid.write(INDENT * 2 + 'assert isinstance(obj, {})\n'.format(classdef.identifier))
        fid.write(INDENT * 2 + 'return {}_to_jsonable(obj, path=path)'.format(
            swagger_to.snake_case(classdef.identifier)))

    fid.write('\n\n')

    fid.write(INDENT + 'raise ValueError("Unexpected `expected` type: {}".format(exp))')


def write_class_to_jsonable(classdef: Classdef, fid: TextIO) -> None:
    fid.write('def {0}_to_jsonable({0}: {1}, path: str = "") -> Dict[str, Any]:\n'.format(
        swagger_to.snake_case(identifier=classdef.identifier), classdef.identifier))

    fid.write(INDENT + '"""Generates a dictionary JSON-able object from an instance of {}."""\n'.format(
        classdef.identifier))

    if not classdef.attributes:
        fid.write(INDENT + 'return dict()')
        return

    fid.write(INDENT + 'res = dict()  # type: Dict[str, Any]\n')
    fid.write('\n')

    variable = swagger_to.snake_case(identifier=classdef.identifier)

    for i, attr in enumerate(classdef.attributes.values()):
        if i > 0:
            fid.write('\n\n')

        if attr.required:
            indent = INDENT
        else:
            fid.write(INDENT + 'if {}.{} is not None:\n'.format(variable, attr.name))
            indent = INDENT * 2

        if isinstance(attr.typedef, (Booldef, Intdef, Floatdef, Strdef)):
            fid.write(indent + 'res["{0}"] = {1}.{0}'.format(attr.name, variable))
        else:
            prefix = indent + 'res["{}"] = to_jsonable('.format(attr.name)
            value = '{}.{}'.format(variable, attr.name)
            expected = '[{}]'.format(expected_type_expression(typedef=attr.typedef))
            path = '"{{}}.{}".format(path))'.format(attr.name)

            line = prefix + ', '.join([value, expected, path])
            if len(line) <= 80:
                fid.write(line)
            else:
                fid.write(prefix + value + ',\n')
                fid.write(' ' * len(prefix) + expected + ',\n')
                fid.write(' ' * len(prefix) + path)

    fid.write('\n')
    fid.write(INDENT + 'return res')


def to_string_expression(typedef: Typedef, expression: str) -> str:
    """
    Wraps the expression in str() if necessary.

    :param typedef: type definition of the variable
    :param expression: to be converted to string in the generated code
    :return: python string expression
    """
    if isinstance(typedef, Strdef):
        return expression

    return 'str({})'.format(expression)


def request_docstring(request: Request) -> str:
    """
    Assembles the docstring of the given request function.

    :param request: function to be documented
    :return: docstring of the request function
    """
    docstring_lines = []  # type: List[str]
    if request.description:
        docstring_lines += request.description.splitlines()
        if request.parameters:
            docstring_lines += ['']

    for param in request.parameters:
        if not param.description:
            docstring_lines += [':param {}:'.format(param.name)]
        else:
            description_lines = param.description.splitlines()
            if len(description_lines) == 1:
                docstring_lines += [':param {}: {}'.format(param.name, param.description)]
            else:
                docstring_lines += [':param {}:'.format(param.name)]
                for line in description_lines:
                    docstring_lines += [INDENT * 2 + line]

    return '\n'.join(docstring_lines)


def write_request(request: Request, fid: TextIO) -> None:
    resp = None  # type: Optional[Response]
    return_type = 'bytes'
    if request.produces == ['application/json']:
        if '200' in request.responses:
            resp = request.responses['200']
            if resp.typedef is not None:
                return_type = type_expression(typedef=resp.typedef, path=request.operation_id + '.' + str(resp.code))
            else:
                # The schema for the response has not been defined. Hence we can not parse the response.
                resp = None
                return_type = 'Any'

    prefix = INDENT + 'def {}('.format(request.operation_id)
    suffix = ') -> {}:'.format(return_type)

    args = ['self']  # type: List[str]
    for param in request.parameters:
        param_type = type_expression(typedef=param.typedef, path=request.operation_id + '.' + param.name)

        if not param.required:
            args.append('{}: Optional[{}] = None'.format(param.name, param_type))
        else:
            args.append('{}: {}'.format(param.name, param_type))

    line = prefix + ', '.join(args) + suffix
    if len(line) <= 80:
        fid.write(line)
    else:
        fid.write(prefix)
        for i, arg in enumerate(args):
            if i > 0:
                fid.write(',\n')
                fid.write(' ' * len(prefix))
            fid.write(arg)

        fid.write(suffix)
    fid.write('\n')

    # assemble the docstring
    docstring = request_docstring(request=request)
    write_docstring(docstring=docstring, indent=INDENT * 2, fid=fid)
    fid.write('\n')

    # path parameters
    name_to_parameters = dict([(param.name, param) for param in request.parameters])

    token_pth = swagger_to.tokenize_path(path=request.path)

    if not token_pth.parameter_to_token_indices:
        fid.write(INDENT * 2 + 'url = self.url_prefix + "{}"'.format(request.path))
    else:
        fid.write(INDENT * 2 + 'url_parts = [self.url_prefix]')
        for i, tkn in enumerate(token_pth.tokens):
            fid.write("\n")

            if i in token_pth.token_index_to_parameter:
                param_name = token_pth.token_index_to_parameter[i]
                param = name_to_parameters[param_name]

                fid.write(
                    INDENT * 2 + 'url_parts.append({})'.format(
                        to_string_expression(typedef=param.typedef, expression=param.name)))
            else:
                fid.write(INDENT * 2 + 'url_parts.append("{}")'.format(
                    tkn.replace('\\', '\\\\').replace('\n', '\\n').replace('\r', '\\r').replace('\t', '\\t')))

        fid.write('\n')
        fid.write(INDENT * 2 + 'url = "".join(url_parts)')

    if request.query_parameters:
        fid.write('\n\n')
        fid.write(INDENT * 2 + 'params = {\n')

        for i, param in enumerate(request.query_parameters):
            if i > 0:
                fid.write(',\n')
            if isinstance(param.typedef, (Booldef, Intdef, Floatdef, Strdef)):
                fid.write(INDENT * 3 + '"{0}": {0}'.format(param.name))
            else:
                fid.write(INDENT * 3 + '"{0}": to_jsonable({0}, expected=[{1}])'.format(
                    param.name, expected_type_expression(typedef=param.typedef)))

        fid.write('}')

    if request.body_parameter and request.formdata_parameters:
        raise KeyError("Both body parameter and form-data parameters are specified. "
                       "The python client does not know how to resolve this request.")

    if request.body_parameter:
        fid.write('\n\n')
        if isinstance(request.body_parameter.typedef, (Booldef, Intdef, Floatdef, Strdef)):
            fid.write(INDENT * 2 + 'data = {}'.format(request.body_parameter.name))
        else:
            fid.write(INDENT * 2 + 'data = to_jsonable({0}, expected=[{1}])'.format(
                request.body_parameter.name,
                expected_type_expression(typedef=request.body_parameter.typedef)))

    if request.formdata_parameters:
        fid.write('\n\n')
        fid.write(INDENT * 2 + 'data = {\n')

        for i, param in enumerate(request.formdata_parameters):
            if i > 0:
                fid.write(',\n')

            if isinstance(param.typedef, (Booldef, Intdef, Floatdef, Strdef)):
                fid.write(INDENT * 3 + '"{0}": {0}'.format(param.name))
            else:
                fid.write(INDENT * 3 + '"{0}": to_jsonable({0}, expected=[{1}])'.format(
                    param.name, expected_type_expression(typedef=param.typedef)))

        fid.write('}')

    if request.file_parameters:
        fid.write('\n\n')
        fid.write(INDENT * 2 + 'files = {\n')

        for i, param in enumerate(request.file_parameters):
            if i > 0:
                fid.write(',\n')

            assert isinstance(param.typedef, Filedef), \
                "Expected parameter {} of path {}.{} to be Filedef, but got: {}".format(
                    param.name, request.path, request.method, type(param.typedef))

            fid.write(INDENT * 3 + '"{0}": {0}'.format(param.name))

        fid.write('}')

    fid.write('\n\n')

    fid.write(INDENT * 2 + 'resp = requests.request(method={!r}, url=url'.format(request.method))
    if request.query_parameters:
        fid.write(', params=params')

    if request.body_parameter:
        fid.write(', json=data')
    elif request.formdata_parameters:
        fid.write(', data=data')
    else:
        # ignore the parameter which we don't know how to handle.
        pass

    if request.file_parameters:
        fid.write(', files=files')

    fid.write(', auth=self.auth)\n')
    fid.write(INDENT * 2 + 'with contextlib.closing(resp):\n')
    fid.write(INDENT * 3 + 'resp.raise_for_status()\n')

    if resp is None:
        if return_type == 'bytes':
            fid.write(INDENT * 3 + 'return resp.content')
        elif return_type == 'Any':
            fid.write(INDENT * 3 + 'return resp.json()')
        else:
            raise NotImplementedError("Unhandled return type of the request {}.{}: {!r}".format(
                request.path, request.method, return_type))
    else:
        fid.write(INDENT * 3 + 'return from_obj(obj=resp.json(), expected=[{}], path="")'.format(
            expected_type_expression(typedef=resp.typedef)))


def write_client(requests: List[Request],
                 fid: TextIO) -> None:
    fid.write("class RemoteCaller:\n")
    fid.write(INDENT + '"""Executes the remote calls to the server."""\n')
    fid.write('\n')

    fid.write(INDENT + 'def __init__(self, url_prefix: str, auth: Optional[requests.auth.AuthBase] = None) -> None:\n')
    fid.write(INDENT * 2 + "self.url_prefix = url_prefix\n")
    fid.write(INDENT * 2 + "self.auth = auth")

    if requests:
        fid.write('\n\n')

    for i, request in enumerate(requests):
        if i > 0:
            fid.write('\n\n')
        write_request(request=request, fid=fid)


def write_client_py(service_name: str,
                    typedefs: MutableMapping[str, Typedef],
                    requests: List[Request],
                    fid: TextIO) -> None:
    """
    Generates the file with the client code.

    :param service_name: used to designate the service that client connects to
    :param typedefs: translated type definitions
    :param requests: translated request functions
    :param fid: target
    :return:
    """
    write_header(service_name=service_name, fid=fid)

    if typedefs:
        classdefs = [typedef for typedef in typedefs.values() if isinstance(typedef, Classdef)]

        if classdefs:
            fid.write('\n\n')
            write_from_obj(classdefs=classdefs, fid=fid)
            fid.write('\n\n\n')
            write_to_jsonable(classdefs=classdefs, fid=fid)
            fid.write('\n\n\n')

        for i, classdef in enumerate(classdefs):
            if i > 0:
                fid.write('\n\n\n')
            write_class(classdef=classdef, fid=fid)
            fid.write('\n\n\n')
            write_class_factory_method(classdef=classdef, fid=fid)
            fid.write('\n\n\n')
            write_class_from_obj(classdef=classdef, fid=fid)
            fid.write('\n\n\n')
            write_class_to_jsonable(classdef=classdef, fid=fid)

    if typedefs:
        fid.write('\n\n\n')

    write_client(requests=requests, fid=fid)

    if requests and typedefs:
        fid.write('\n\n\n')

    write_footer(fid=fid)
    fid.write('\n')



"""
Parses Swagger specification and generates server and client stubs.
"""
import collections
import re
import string
from typing import List, MutableMapping

# pylint: disable=missing-docstring


def parse_definition_ref(ref: str) -> str:
    prefix = '#/definitions/'
    if not ref.startswith(prefix):
        raise ValueError("Expected a ref with prefix {!r}, but got: {!r}".format(prefix, ref))

    return ref[len(prefix):]


def parse_parameter_ref(ref: str) -> str:
    prefix = '#/parameters/'
    if not ref.startswith(prefix):
        raise ValueError("Expected a ref with prefix {!r}, but got: {!r}".format(prefix, ref))

    return ref[len(prefix):]


# abbreviations to be treaded specially in snake_case and camelCase conversions
SPECIALS = ['URLs', 'IDs', 'URL', 'ID', 'HTTP', 'HTTPS']


def camel_case_split(identifier: str) -> List[str]:
    if identifier == '':
        raise ValueError("Unexpected empty identifier")

    cur = 0

    parts = []  # type: List[str]
    while cur < len(identifier):
        found_special = False
        for special in SPECIALS:
            if identifier[cur:cur + len(special)] == special:
                parts.append(special)
                cur += len(special)
                found_special = True

        if not found_special:
            if identifier[cur] in string.ascii_uppercase:
                parts.append('')
                parts[-1] += identifier[cur]
                cur += 1
            else:
                if len(parts) == 0:
                    parts.append('')

                parts[-1] += identifier[cur]
                cur += 1

    return parts


def capital_camel_case(identifier: str) -> str:
    """
    Translates an arbitrary identifier to a CamelCase.

    :param identifier: arbitrary identifier
    :return: identifier as CamelCase
    """
    if identifier == '':
        raise ValueError("Unexpected empty identifier")

    parts = identifier.split("_")

    camel_parts = []  # type: List[str]
    for part in parts:
        camel_parts.extend(camel_case_split(identifier=part))

    new_parts = []  # type: List[str]
    for part in camel_parts:
        part = part.lower()

        if part in ['url', 'id']:
            new_parts.append(part.upper())
        elif part == 'urls':
            new_parts.append('URLs')
        elif part == 'ids':
            new_parts.append('IDs')
        else:
            new_parts.append(part[0].upper() + part[1:].lower())

    return "".join(new_parts)


def camel_case(identifier: str) -> str:
    """
    Translates an arbitrary identifier to a camelCase.

    :param identifier: arbitrary identifier
    :return: identifier as camelCase
    """
    if identifier == '':
        raise ValueError("Unexpected empty identifier")

    parts = identifier.split("_")

    camel_parts = []  # type: List[str]
    for part in parts:
        camel_parts.extend(camel_case_split(identifier=part))

    new_parts = [camel_parts[0].lower()]  # type: List[str]

    for part in camel_parts[1:]:
        part = part.lower()

        if part in ['url', 'id']:
            new_parts.append(part.upper())
        elif part == 'urls':
            new_parts.append('URLs')
        elif part == 'ids':
            new_parts.append('IDs')
        else:
            new_parts.append(part[0].upper() + part[1:].lower())

    return "".join(new_parts)


def snake_case(identifier: str) -> str:
    """
    Converts an indentifier to a lowercase snake case.

    :param identifier: to be converted
    :return: lowercase snake_case identifier
    """
    if identifier == '':
        return ''

    parts = camel_case_split(identifier=identifier)

    result = '_'.join(parts)
    return result.lower()


class TokenizedPath:
    def __init__(self):
        self.tokens = []  # type: List[str]
        self.parameter_to_token_indices = collections.OrderedDict()  # type: MutableMapping[str, List[int]]
        self.token_index_to_parameter = collections.OrderedDict()  # type: MutableMapping[int, str]


PATH_TOKENIZATION_RE = re.compile(r'\{(?P<name>[a-zA-Z0-9_]*)\}|[^{]+|.')


def tokenize_path(path: str) -> TokenizedPath:
    """
    Tokenizes the path coming from a Swagger spec to a dictionary such that you can modify the path parameters
    easily.

    :param path: original path
    :return: tokenized path with the dictionary
    """
    token_pth = TokenizedPath()

    for mtch in PATH_TOKENIZATION_RE.finditer(path):
        token_pth.tokens.append(mtch.group(0))

        if mtch.group("name") is not None:
            name = mtch.group("name")
            if name not in token_pth.parameter_to_token_indices:
                token_pth.parameter_to_token_indices[name] = []

            token_pth.parameter_to_token_indices[name].append(len(token_pth.tokens) - 1)
            token_pth.token_index_to_parameter[len(token_pth.tokens) - 1] = name

    return token_pth



#!/usr/bin/env python3
"""
Generates code for an TypeScript client with Angular5.
"""

import collections
from typing import MutableMapping, Optional, TextIO, List  # pylint: disable=unused-import

import swagger_to
import swagger_to.intermediate
import swagger_to.swagger

INDENT = ' ' * 4


class Typedef:
    """
    Represents a typescript type.
    """

    def __init__(self) -> None:
        self.description = ''
        self.identifier = ''

    def __str__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, self.identifier)


class Booleandef(Typedef):
    """
    Represents typescript booleans.
    """
    pass


class Numberdef(Typedef):
    """
    Represents typescript numbers.
    """
    pass


class Stringdef(Typedef):
    """
    Represents typescript strings.
    """
    pass


class Arraydef(Typedef):
    """
    Represents typescript arrays.
    """

    def __init__(self) -> None:
        self.items = None  # type: Optional[Typedef]
        super().__init__()

    def __str__(self) -> str:
        return '{}({}, items: {})'.format(self.__class__.__name__, self.identifier, self.items)


class Mapdef(Typedef):
    """
    Represents typescript maps.
    """

    def __init__(self) -> None:
        self.values = None  # type: Optional[Typedef]
        super().__init__()

    def __str__(self) -> str:
        return '{}({}, values: {})'.format(self.__class__.__name__, self.identifier, self.values)


class Property:
    """
    Represents a typescript class property.
    """

    def __init__(self) -> None:
        self.name = ''
        self.description = ''
        self.typedef = None  # type: Optional[Typedef]
        self.required = False

    def __str__(self) -> str:
        return '{}: {}'.format(self.name, self.typedef)


class Classdef(Typedef):
    """
    Represents typescript classes.
    """

    def __init__(self) -> None:
        self.properties = collections.OrderedDict()  # type: MutableMapping[str, Property]
        super().__init__()

    def __str__(self) -> str:
        return '{}({}, properties: {})'.format(self.__class__.__name__, self.identifier,
                                               ', '.join([str(val) for val in self.properties.values()]))


class Parameter:
    """
    Represents a parameter of a request.
    """

    def __init__(self) -> None:
        self.name = ''
        self.typedef = None  # type: Optional[Typedef]
        self.required = False


class Response:
    """
    Represents a response from a request.
    """

    def __init__(self) -> None:
        self.code = ''
        self.typedef = None  # type: Optional[Typedef]
        self.description = ''


class Request:
    """
    Represents a request function of the client.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self) -> None:
        self.operation_id = ''
        self.path = ''
        self.method = ''
        self.description = ''

        # the union of all parameters
        self.parameters = []  # type: List[Parameter]

        # parameters split into categories
        self.body_parameter = None  # type: Optional[Parameter]
        self.query_parameters = []  # type: List[Parameter]
        self.path_parameters = []  # type: List[Parameter]

        self.responses = collections.OrderedDict()  # type: MutableMapping[str, Response]


def to_typedef(intermediate_typedef: swagger_to.intermediate.Typedef) -> Typedef:
    """
    Translates the intermediate typedef to a typescript typedef.

    :param intermediate_typedef: to be translated
    :return:
    """
    typedef = None  # type: Optional[Typedef]

    if isinstance(intermediate_typedef, swagger_to.intermediate.Primitivedef):
        if intermediate_typedef.type == 'boolean':
            typedef = Booleandef()
        elif intermediate_typedef.type == 'number' or intermediate_typedef.type == 'integer':
            typedef = Numberdef()
        elif intermediate_typedef.type == 'string':
            typedef = Stringdef()
        else:
            raise NotImplementedError("Converting intermediate type to Typescript is not supported: {}".format(
                intermediate_typedef.type))

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Arraydef):
        typedef = Arraydef()
        typedef.items = to_typedef(intermediate_typedef=intermediate_typedef.items)

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Mapdef):
        typedef = Mapdef()
        typedef.values = to_typedef(intermediate_typedef=intermediate_typedef.values)

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Objectdef):
        typedef = Classdef()

        for intermediate_prop in intermediate_typedef.properties.values():
            prop = Property()
            prop.description = intermediate_prop.description
            prop.name = intermediate_prop.name
            prop.typedef = to_typedef(intermediate_typedef=intermediate_prop.typedef)
            prop.required = intermediate_prop.required

            typedef.properties[prop.name] = prop
    else:
        raise NotImplementedError("Converting intermediate typedef to Typescript is not supported: {!r}".format(
            type(intermediate_typedef)))

    typedef.description = intermediate_typedef.description
    typedef.identifier = intermediate_typedef.identifier

    assert typedef is not None

    return typedef


def to_typedefs(
        intermediate_typedefs: MutableMapping[str, swagger_to.intermediate.Typedef]) -> MutableMapping[str, Typedef]:
    """
    Translates the intermediate typedefs to typescript typedefs.

    :param intermediate_typedefs: to be translated
    :return: translated typedefs
    """
    typedefs = collections.OrderedDict()  # type: MutableMapping[str, Typedef]

    for intermediate_typedef in intermediate_typedefs.values():
        assert intermediate_typedef is not None

        typedef = to_typedef(intermediate_typedef=intermediate_typedef)
        typedefs[typedef.identifier] = typedef

    return typedefs


def anonymous_or_get_typedef(intermediate_typedef: swagger_to.intermediate.Typedef,
                             typedefs: MutableMapping[str, Typedef]) -> Typedef:
    """
    If the type has an identifier, it is retrieved from the translated typedefs.

    Otherwise, it is translated on the spot to the corresponding typescript type.

    :param intermediate_typedef: to be translated
    :param typedefs: translated type definitions
    :return: translated type
    """
    if intermediate_typedef.identifier:
        if intermediate_typedef.identifier not in typedefs:
            raise KeyError("Intermediate typedef not found in the translated typedefs: {!r}".format(
                intermediate_typedef.identifier))

        return typedefs[intermediate_typedef.identifier]

    return to_typedef(intermediate_typedef=intermediate_typedef)


def to_parameter(intermediate_parameter: swagger_to.intermediate.Parameter,
                 typedefs: MutableMapping[str, Typedef]) -> Parameter:
    """
    Translates an intermediate parameter to a typescript parameter.

    :param intermediate_parameter: to be translated
    :param typedefs: translated type definitions
    :return: translated parameter
    """
    param = Parameter()
    param.name = intermediate_parameter.name
    param.typedef = anonymous_or_get_typedef(intermediate_typedef=intermediate_parameter.typedef, typedefs=typedefs)
    param.required = intermediate_parameter.required
    return param


def to_response(intermediate_response: swagger_to.intermediate.Response,
                typedefs: MutableMapping[str, Typedef]) -> Response:
    """
    Translates an intermediate response to a typescript response.

    :param intermediate_response: to be translated
    :param typedefs: translated type definitions
    :return: translated response
    """
    resp = Response()
    resp.code = intermediate_response.code
    resp.typedef = None if intermediate_response.typedef is None else \
        anonymous_or_get_typedef(intermediate_typedef=intermediate_response.typedef, typedefs=typedefs)
    resp.description = intermediate_response.description
    return resp


def to_request(endpoint: swagger_to.intermediate.Endpoint, typedefs: MutableMapping[str, Typedef]) -> Request:
    """
    Translates an endpoint to a client request function.

    :param endpoint: to be translated
    :param typedefs: translated type definitions
    :return: request function
    """
    if endpoint.produces != ['application/json']:
        raise ValueError("Can not translate an end point to typescript client "
                         "which does not produces strictly application/json: {} {}".format(
                             endpoint.path, endpoint.method))

    req = Request()
    req.description = endpoint.description
    req.method = endpoint.method
    req.operation_id = endpoint.operation_id
    req.path = endpoint.path

    for intermediate_param in endpoint.parameters:
        param = to_parameter(intermediate_parameter=intermediate_param, typedefs=typedefs)

        if intermediate_param.in_what == 'body':
            if req.body_parameter is not None:
                raise KeyError("Duplicate body parameters in an endpoint: {} {}".format(
                    req.body_parameter.name, intermediate_param.name))

            req.body_parameter = param
        elif intermediate_param.in_what == 'query':
            req.query_parameters.append(param)
        elif intermediate_param.in_what == 'path':
            req.path_parameters.append(param)
        else:
            raise NotImplementedError("Unsupported parameter 'in' to Typescript translation: {}".format(
                intermediate_param.in_what))

        req.parameters.append(param)

    # parameters are sorted so that first come the required ones; typescript requires the optional ones to come
    # at the end.
    req.parameters.sort(key=lambda param: not param.required)

    for code, intermediate_resp in endpoint.responses.items():
        req.responses[code] = to_response(intermediate_response=intermediate_resp, typedefs=typedefs)

    return req


def to_requests(endpoints: List[swagger_to.intermediate.Endpoint],
                typedefs: MutableMapping[str, Typedef]) -> List[Request]:
    """
    Translates the endpoints to typescript request functions.

    Endpoints which do not produce strictly 'application/json' are ignored.

    :param endpoints: to be translated
    :param typedefs: translated type definitions
    :return: translated request functions
    """
    requests = []  # type: List[Request]
    for endpoint in endpoints:
        if endpoint.produces != ['application/json']:
            continue

        requests.append(to_request(endpoint=endpoint, typedefs=typedefs))

    return requests


def write_header(fid: TextIO) -> None:
    """
    Writes the header (same for all the typescript clients).

    :param fid: target
    :return:
    """
    fid.write("// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!\n\n")

    fid.write("import { Injectable } from '@angular/core';\n"
              "import { Http } from '@angular/http';\n"
              "import { HttpErrorResponse } from '@angular/common/http';\n"
              "import { Observable } from 'rxjs/Rx';\n"
              "import { RequestOptions } from '@angular/http';\n\n")


def write_footer(fid: TextIO) -> None:
    """
    Writes the footer (same for all the typescript clients).

    :param fid: target
    :return:
    """

    fid.write("\n\n// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!\n")


def type_expression(typedef: Typedef, path: Optional[str] = None) -> str:
    """
    Translates the typedef to a type expression.

    :param typedef: to be translated
    :param path: path in the intermediate representation
    :return: type expression
    """
    if isinstance(typedef, Booleandef):
        return 'boolean'
    elif isinstance(typedef, Numberdef):
        return 'number'
    elif isinstance(typedef, Stringdef):
        return 'string'
    elif isinstance(typedef, Arraydef):
        return 'Array<' + type_expression(typedef=typedef.items, path=str(path) + '.items') + ">"
    elif isinstance(typedef, Mapdef):
        return 'Map<string, ' + type_expression(typedef=typedef.values, path=str(path) + '.values') + ">"
    elif isinstance(typedef, Classdef):
        if typedef.identifier == '':
            raise NotImplementedError(
                "Translating an anonymous class to a Typescript type expression is not supported: {}".format(path))

        return typedef.identifier
    else:
        raise NotImplementedError("Translating the typedef to a type expression is not supported: {!r}: {}".format(
            type(typedef), path))


def write_description(description: str, indent: str, fid: TextIO) -> None:
    """
    Writes a description as // comment block.
    :param description: to be written
    :param indent: indention as prefix to each line
    :param fid: target
    :return:
    """
    lines = description.strip().splitlines()
    for i, line in enumerate(lines):
        rstripped = line.rstrip()
        if len(rstripped) > 0:
            fid.write(indent + '// {}'.format(rstripped))
        else:
            fid.write(indent + '//')

        if i < len(lines) - 1:
            fid.write('\n')


def write_type_definition(typedef: Typedef, fid: TextIO) -> None:
    """
    Writes the type definition in the Typescript code.

    :param typedef: to be declared
    :param fid: target
    :return:
    """
    if typedef.identifier == '':
        raise ValueError("Expected a typedef with an identifier, but got a typedef with an empty identifier.")

    if typedef.description:
        write_description(description=typedef.description, indent='', fid=fid)
        fid.write('\n')

    if isinstance(typedef, Classdef):
        fid.write("export interface {} {{\n".format(typedef.identifier))
        for i, prop in enumerate(typedef.properties.values()):
            if i > 0:
                fid.write('\n')

            if prop.description:
                write_description(description=prop.description, indent=INDENT, fid=fid)
                fid.write("\n")

            type_expr = type_expression(typedef=prop.typedef, path='{}.{}'.format(typedef.identifier, prop.name))

            if not prop.required:
                fid.write(INDENT + '{}?: {};\n'.format(prop.name, type_expr))
            else:
                fid.write(INDENT + '{}: {};\n'.format(prop.name, type_expr))

        fid.write("}")
    else:
        fid.write("type {} = {};".format(typedef.identifier, type_expression(typedef=typedef, path=typedef.identifier)))


def write_type_definitions(typedefs: MutableMapping[str, Typedef], fid: TextIO) -> None:
    """
    Writes all type definitions as Typescript code.

    :param typedefs: type definitions to be declared
    :param fid: target
    :return:
    """
    for i, typedef in enumerate(typedefs.values()):
        if i > 0:
            fid.write('\n\n')

        write_type_definition(typedef=typedef, fid=fid)


def to_string_expression(typedef: Typedef, variable: str) -> str:
    """
    Wraps the variable with .toString() if necessary.

    :param typedef: type definition of the variable
    :param variable: to be converted to string in the generated code
    :return: typescript expression
    """
    if isinstance(typedef, Stringdef):
        return variable

    return '{}.toString()'.format(variable)


def write_request(request: Request, fid: TextIO) -> None:
    """
    Generates the code of the request function.

    :param request: function definition
    :param fid: target
    :return:
    """
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements
    # pylint: disable=too-many-branches
    description = 'Sends a request to the endpoint: {} {}'.format(request.path, request.method)
    if request.description:
        description += '\n\n' + request.description
    write_description(description, INDENT, fid)
    fid.write('\n')

    prefix = INDENT + 'public {}('.format(request.operation_id)

    args = []  # type: List[str]
    for param in request.parameters:
        args.append('{}{}: {}'.format(
            param.name, '?' if not param.required else '', type_expression(typedef=param.typedef)))

    return_type = ''
    if '200' in request.responses:
        resp = request.responses['200']
        if resp.typedef is not None:
            return_type = type_expression(typedef=resp.typedef)

    if not return_type:
        return_type = 'any'

    suffix = '): Observable<{} | HttpErrorResponse> {{'.format(return_type)

    line = prefix + ', '.join(args) + suffix
    if len(line) <= 120:
        fid.write(line)
    else:
        fid.write(prefix)
        fid.write('\n')

        for i, arg in enumerate(args):
            fid.write(3 * INDENT + arg)

            if i < len(args) - 1:
                fid.write(',\n')

        fid.write(suffix)
    fid.write('\n')

    name_to_parameters = dict([(param.name, param) for param in request.parameters])

    rel_path = request.path[1:] if request.path.startswith('/') else request.path

    # path parameters
    token_pth = swagger_to.tokenize_path(path=rel_path)

    if not token_pth.parameter_to_token_indices and not request.query_parameters:
        fid.write(INDENT * 2 + 'const url = this.url_prefix + "{}";\n'.format(rel_path))
    else:
        if not token_pth.parameter_to_token_indices:
            fid.write(INDENT * 2 + 'let url = this.url_prefix + "{}";'.format(rel_path))
        else:
            fid.write(INDENT * 2 + 'let url = this.url_prefix;')
            for i, tkn in enumerate(token_pth.tokens):
                fid.write("\n")

                if i in token_pth.token_index_to_parameter:
                    param_name = token_pth.token_index_to_parameter[i]
                    param = name_to_parameters[param_name]

                    fid.write(INDENT * 2 + 'url += encodeURIComponent({});'.format(
                        to_string_expression(typedef=param.typedef, variable=param.name)))
                else:
                    fid.write(INDENT * 2 + 'url += encodeURIComponent("{}");'.format(
                        tkn.replace('\\', '\\\\').replace('\n', '\\n').replace('\r', '\\r').replace('\t', '\\t')))

    if request.path_parameters and request.query_parameters:
        fid.write("\n")

    # query parameters
    if request.query_parameters:
        fid.write('\n')
        fid.write(INDENT * 2 + 'url += "?";\n')

        for i, param in enumerate(request.query_parameters):
            amp = ''  # used to concatenate query parameters
            if i > 0:
                fid.write("\n\n")
                amp = '&'

            if param.required:
                fid.write(INDENT * 2 + 'url += "{}{}=" + encodeURIComponent({});'.format(
                    amp, param.name, to_string_expression(typedef=param.typedef, variable=param.name)))
            else:
                fid.write(INDENT * 2 + 'if ({}) {{\n'.format(param.name))
                fid.write(INDENT * 3 + 'url += "{}{}=" + encodeURIComponent({});\n'.format(
                    amp, param.name, to_string_expression(typedef=param.typedef, variable=param.name)))
                fid.write(INDENT * 2 + '}')

        fid.write('\n')

    fid.write('\n')

    mth = request.method.lower()
    if request.body_parameter is not None:
        if request.body_parameter.required:
            fid.write(INDENT * 2 + 'let observable = this.http.request(url, \n')
            fid.write(INDENT * 3 + 'new RequestOptions({{method: "{0}", body: JSON.stringify({1})}}));\n'.format(
                mth, request.body_parameter.name))
        else:
            fid.write(INDENT * 2 + 'let observable: Observable<any>;\n')
            fid.write(INDENT * 2 + 'if ({}) {{\n'.format(request.body_parameter.name))
            fid.write(INDENT * 3 + 'this.http.request(url, \n')
            fid.write(INDENT * 4 + 'new RequestOptions({{method: "{0}", body: JSON.stringify({1})}}));\n'.format(
                mth, request.body_parameter.name))
            fid.write(INDENT * 2 + '} else {\n')
            fid.write(INDENT * 3 + 'observable = this.http.request(url, '
                      'new RequestOptions({{method: "{0}"}}));\n'.format(mth))
            fid.write(INDENT * 2 + '}\n')
    else:
        fid.write(INDENT * 2 + 'let observable = this.http.{}(url);\n'.format(mth))

    return_var = 'observable'
    if return_type != 'any':
        fid.write(
            INDENT * 2 + 'let typed_observable = observable.map(res => (res.json() as {}));\n'.format(return_type))
        return_var = 'typed_observable'

    fid.write(INDENT * 2 + 'if (this.on_error) {\n')
    fid.write(INDENT * 3 + 'return {0}.catch(err => this.on_error(err))\n'.format(return_var))
    fid.write(INDENT * 2 + '}\n')
    fid.write(INDENT * 2 + 'return {};\n'.format(return_var))

    fid.write(INDENT + '}')


def write_client(requests: List[Request], fid: TextIO) -> None:
    """
    Generates the client.

    :param requests: translated request functions
    :param fid: target
    :return:
    """
    fid.write("@Injectable()\n")
    fid.write("export class RemoteCaller {\n")
    fid.write(INDENT + "public url_prefix: string;\n")
    fid.write(INDENT + "public on_error: (error: HttpErrorResponse) => Observable<HttpErrorResponse> | null;\n")
    fid.write("\n")
    fid.write(INDENT + "constructor(private http: Http) {\n")
    fid.write(INDENT * 2 + 'this.url_prefix = "";\n')
    fid.write(INDENT * 2 + 'this.on_error = null;\n')
    fid.write(INDENT + '}\n\n')
    fid.write(INDENT + "public set_url_prefix(url_prefix: string) {\n")
    fid.write(INDENT * 2 + "this.url_prefix = url_prefix;\n")
    fid.write(INDENT + '}\n\n')
    fid.write(INDENT + "public set_on_error("
              "on_error: (error: HttpErrorResponse) => Observable<HttpErrorResponse> | null) {\n")
    fid.write(INDENT * 2 + "this.on_error = on_error;\n")
    fid.write(INDENT + '}')

    for request in requests:
        fid.write('\n\n')
        write_request(request=request, fid=fid)

    fid.write("\n}")


def write_client_ts(typedefs: MutableMapping[str, Typedef], requests: List[Request], fid: TextIO) -> None:
    """
    Generates the file with the client code.

    :param typedefs: translated type definitions
    :param requests: translated request functions
    :param fid: target
    :return:
    """
    write_header(fid=fid)

    if typedefs:
        write_type_definitions(typedefs=typedefs, fid=fid)

    if requests and typedefs:
        fid.write('\n\n')

    if requests:
        write_client(requests=requests, fid=fid)
        write_footer(fid=fid)



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""
import os

from setuptools import setup, find_packages

# pylint: disable=redefined-builtin

here = os.path.abspath(os.path.dirname(__file__))  # pylint: disable=invalid-name

with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()  # pylint: disable=invalid-name

setup(
    name='swagger_to',
    version='2.2.3',
    description='Generates code from swagger definitions; an alternative to swagger codegen',
    long_description=long_description,
    url='https://github.com/Parquery/swagger-to',
    author='Marko Ristin',
    author_email='marko@parquery.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
    ],
    license="License :: OSI Approved :: MIT License",
    keywords='swagger code generation python elm go typescript server client angular',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    install_requires=['pyyaml>=3.12'],
    extras_require={'dev': ['mypy==0.560', 'pylint==1.8.2', 'yapf==0.20.2']},
    py_modules=['swagger_to'],
    package_data={"swagger_to": ["py.typed"]},
    scripts=[
        'bin/swagger_to_go_server.py', 'bin/swagger_to_py_client.py', 'bin/swagger_to_ts_angular5_client.py',
        'bin/swagger_to_elm_client.py', 'bin/swagger_style.py'
    ],
)



#!/usr/bin/env python3
"""
Reads a correct swagger file and produces Go code
"""
import argparse
import os

import swagger_to.go_server
import swagger_to.intermediate
import swagger_to.py_client
import swagger_to.swagger


def main() -> None:
    """"
    Main routine
    """
    parser = argparse.ArgumentParser("Reads a correct swagger file and produces python client code")
    parser.add_argument("--swagger_path", help="path to the swagger file", required=True)
    parser.add_argument("--outpath", help="path to the output file", required=True)
    parser.add_argument("--force", help="overwrite existing files", action="store_true")
    args = parser.parse_args()

    swagger_path = str(args.swagger_path)
    out_path = str(args.outpath)
    force = bool(args.force)

    if not os.path.exists(args.swagger_path):
        raise FileNotFoundError("Swagger file does not exist: {}".format(swagger_path))

    os.makedirs(os.path.dirname(out_path), exist_ok=True)

    if not force and os.path.exists(out_path):
        raise FileExistsError("Output path already exists and --force was not specified: {}".format(out_path))

    swagger, errs = swagger_to.swagger.parse_yaml_file(path=swagger_path)
    if errs:
        raise AssertionError("Errors in {}:\n{}".format(swagger_path, "\n".join(errs)))

    intermediate_typedefs = swagger_to.intermediate.to_typedefs(swagger=swagger)
    intermediate_params = swagger_to.intermediate.to_parameters(swagger=swagger, typedefs=intermediate_typedefs)

    endpoints = swagger_to.intermediate.to_endpoints(
        swagger=swagger, typedefs=intermediate_typedefs, params=intermediate_params)

    py_typedefs = swagger_to.py_client.to_typedefs(intermediate_typedefs=intermediate_typedefs)

    if 'RemoteCaller' in py_typedefs:
        raise ValueError("A definition was specified in the swagger with the name 'RemoteCaller', "
                         "but it's reserved for the Python client class.")

    py_requests = swagger_to.py_client.to_requests(endpoints=endpoints, typedefs=py_typedefs)

    with open(out_path, 'wt') as fid:
        swagger_to.py_client.write_client_py(
            service_name=swagger.name, typedefs=py_typedefs, requests=py_requests, fid=fid)

    print("Generated python client code in: {}".format(out_path))


if __name__ == "__main__":
    main()
